const fs = require('fs')
const path = require('path')
const {convertNodeJS, finish} = require('./node')
const {simpleConvertValueToString} = require('./browser/misc/ast-tools')
const {throwError} = require('./browser/misc/packet-builder')

const argv = require('yargs')
.strict(true)
.option('css', {
  demandOption: false,
  type: 'string',
  describe: 'Input CSS file'
})
.option('html', {
  demandOption: true,
  type: 'string',
  describe: 'Input (X)HTML file'
})
.option('output', {
  demandOption: true,
  type: 'string',
  describe: 'Output (X)HTML file'
})
.option('nocssmap', {
  type: 'boolean',
  describe: 'Show warnings/errors with the location in the CSS file, not the original SASS/LESS file'
})
.option('noprogress', {
  type: 'boolean',
  describe: 'Do not show the progress bars'
})
.option('diffmodeclassnames', {
  type: 'boolean',
  describe: 'Collapse the set of autogenerated class names for easier regression testing at the expense of providing more precise source mapping. It also strips any unused class names from the resulting HTML file so only the autogenerated ones remain'
})
.option('verbose', {
  type: 'boolean',
  describe: 'Enable verbose logging'
})
.option('debug', {
  type: 'boolean',
  describe: 'Show elapsed time rather than the ETA'
})
.option('timeout', {
  type: 'number',
  default: 60,
  describe: 'Specify the timeout for opening a page (in seconds). Set to 0 to disable'
})
.option('nostrict', {
  type: 'boolean',
  describe: 'Offer more lenient processing (`:before` is OK, invalid images loaded via `url(...)` are OK)'
})
.argv

let cssPath = argv.css
let htmlPath = argv.html
let htmlOutputPath = argv.output


if (!htmlPath) { console.error('Missing HTML input file'); process.exit(1) }
if (!htmlOutputPath) { console.error('Missing HTML output file'); process.exit(1) }

if (argv.diffmodeclassnames) {
  console.warn('WARNING: The autogenerated class names will not match their source. They will be collapsed so that diffing is easier.')
}

if (cssPath) {
  cssPath = path.resolve(process.cwd(), cssPath)
}
htmlPath = path.resolve(process.cwd(), htmlPath)
htmlOutputPath = path.resolve(process.cwd(), htmlOutputPath)

const htmlOutputLcovPath = `${htmlOutputPath}.lcov`
const htmlOutputSourceMapPath = `${htmlOutputPath}.map`
const htmlOutputVanillaCSSPath = `${htmlOutputPath}.css`
const htmlOutputSourceMapFilename = path.basename(htmlOutputSourceMapPath)


function coverageDataToLcov(htmlOutputPath, coverageData) {
  const lines = []

  for (const filePath in coverageData) {
    // LCOV files MUST be absolute paths (or genhtml will break)
    const absoluteFilePath = path.resolve(path.dirname(htmlOutputPath), filePath) // path.resolve(filePath)
    const countData = coverageData[filePath]
    // SF:./rulesets/output/biology.css
    lines.push(`SF:${absoluteFilePath}`)
    for (const key in countData) {
      const {count, start, end} = countData[key]
      lines.push(`DA:${start.line},${count}`)
    }
    lines.push(`end_of_record`)
  }

  return lines.join('\n')
}


convertNodeJS(cssPath, htmlPath, htmlOutputPath, argv)
.then(({html: actualOutput, sourceMap, coverageData}) => {
  fs.writeFileSync(htmlOutputPath, actualOutput)
  fs.writeFileSync(htmlOutputSourceMapPath, sourceMap)
  fs.writeFileSync(htmlOutputLcovPath, coverageDataToLcov(htmlOutputPath, coverageData))
  return finish() // Close the browser so this process exits
}).catch((err) => {
  console.error(err)
  process.exit(111)
})
